"""
 A Directed Graph Object
 
 Author: Ben Kybartas
 Student ID: 260477933
 Date: April 10, 2012
"""

from Node import *
from Edge import *

from numpy import *

class Graph:
	
	"""The initialization function
	
	Args:
		name (str): The name of our graph
		nodes (array): Our starting node
	"""
	def __init__(self, name, nodes):
		self._name = name
		self._nodes = nodes
		self._edges = []

		self._adjacency = zeros((len(nodes), len(nodes)))
		
		self._player = None
		
	"""Get the name of our graph

	Returns:
		(str): The name of the graph
	"""
	def get_name(self):
		return self._name
	
	"""Set our player"""
	def set_player(self, player):
		self._player = player
	
	"""Get our player"""
	def get_player(self):
		return self._player
		
	"""Get the adjacency graph"""
	def get_adjacency(self):
		return self._adjacency
	
	"""Set the adjacency between two nodes
	
	Params:
		from_node (Node): The node from which the adjacency occurs
		to_node (Node): The node to which the adjacency ends
	"""
	def setAdj(self, from_node, to_node):
		i = self._nodes.index(from_node)
		j = self._nodes.index(to_node)
		
		#Note that adjacency is one, regardless of number
		#of edges leading to or from it.
		self._adjacency[i][j] = 1

	"""Update our adjacency graph
	
	Used if we've resized the graph recently
	"""
	def update_adj(self):
		for row in range(len(self._adjacency)):
			for col in range(len(self._adjacency)):
				
				#Check for a connection
				connected, edge = self._nodes[row].connected_to(self._nodes[col])
				if connected:
					
					#If so set an adjacency
					self.setAdj(self._nodes[row], self._nodes[col])
					
	"""Set the name of our graph

	Args:
		new_name(str): The new name of the graph
	"""
	def set_name(self, new_name):
		self._name = new_name	

	"""Set the node name"""
	def set_node_name(self, node, new_name):
		num = self._nodes.index(node)
		self._nodes[num].set_name(new_name)

	"""Set the node attributes"""
	def set_node_attribute(self, node, attribute_name, new_value):
		num = self._nodes.index(node)
		self._nodes[num].modify_attribute(attribute_name, new_value)
	
	"""Get the graph edges"""
	def get_edges(self):
		return self._edges
	
	"""Get the graph nodes"""
	def get_nodes(self):
		return self._nodes
		
	"""Add a node to the graph
	
	Params:
		node (Node): The node to be added
	"""
	def add_node(self, node):
		self._nodes.append(node)
		
		#Update our adjacencies
		new_adj = zeros((len(self._nodes), len(self._nodes)))
		for row in range(len(self._adjacency)):
			for col in range(len(self._adjacency)):
				new_adj[row][col] = self._adjacency[row][col]
		self._adjacency = new_adj
			
	"""Add an edge to the graph"""
	def add_edge(self, edge):
		self._edges.append(edge)
	
	"""Get a node from the id"""
	def get_node_from_id(self, node_id):
		return self._nodes[node_id]
	
	"""Modify an edge in the graph
	
	Set the node going from from_node to to_node
	to have a new name given by new_edge
	"""
	def modify_edge(self, from_node, to_node, new_edge):
		exists, edge = from_node.connected_to(to_node)
		if exists:
			edge.set_name(new_edge)
			
	"""Delete edges in the graph"""
	def delete_edges(self, edges):
		for edge in edges:
			self._edges.remove(edge)
	
	"""Add edges in the graph"""
	def add_edges(self, edges):
		self._edges.extend(edges)
			
			
	""" -------------------------------------------
		
		
		THE SUBGRAPH ISOMORPHISM ALGORITHM!


		Here we are using the Ullman algorithm for
		subgraph isomorphism
	---------------------------------------------"""
	
	#Our main equation which gets called
	def contains_subgraph(self, s_graph):

		#We use pseudo case statements 
		#to define where we need to go in
		#an algorithm next
		next_step = 1
		
		while not next_step == -1:
			if next_step == 1:
				next_step = self.step_1(s_graph)
			elif next_step == 2:
				next_step = self.step_2()
			elif next_step == 3:
				next_step = self.step_3()
			elif next_step == 4:
				next_step = self.step_4()
			elif next_step == 5:
				next_step = self.step_5()
			elif next_step == 6:
				next_step = self.step_6()
			elif next_step == 7:
				next_step = self.step_7()


		#Set our matches array
		return_array = []

		for match in self._matches:
			new_array = []
			for row in range(self._p_a):
				for col in range(self._p_b):
					if match[row][col] == 1:
						new_array.append(self._nodes[col])
			return_array.append(new_array)
			
		#Refine our return array
		finalized_return_array = []

		#
		# What we are fundamentally doing below, is to ensure the edges
		# in the return array are the same as the ones in our sub-graph
		#
		for array in return_array:
			Good = True
			for node in range(len(array) - 1):
				
				#Check to outgoing edges
				connected, edge = s_graph.get_nodes()[node].connected_to(s_graph.get_nodes()[node + 1])
				if connected:
					connected_2, edge_2 = array[node].connected_to(array[node + 1])
					if (not connected_2) or (not edge_2.name_equals(edge.get_name())):
						Good = False
						
				#Check the incoming edges
				connected, edge = s_graph.get_nodes()[node].connected_from(s_graph.get_nodes()[node + 1])
				if connected:
					connected_2, edge_2 = array[node].connected_from(array[node + 1])
					if (not connected_2) or (not edge_2.name_equals(edge.get_name())):
						Good = False
			
			#Additionally, check the first node to the last node (which wasn't done above)
			#Check the outgoing edges
			connected, edge = s_graph.get_nodes()[node + 1].connected_to(s_graph.get_nodes()[0])
			if connected:
				connected_2, edge_2 = array[node + 1].connected_to(array[0])
				if (not connected_2) or (not edge_2.name_equals(edge.get_name())):
					Good = False
			
			#Check the incoming edges
			connected, edge = s_graph.get_nodes()[node + 1].connected_from(s_graph.get_nodes()[0])
			if connected:
				connected_2, edge_2 = array[node + 1].connected_from(array[0])
				if (not connected_2) or (not edge_2.name_equals(edge.get_name())):
					Good = False
					
			#If its good, attach it to our finalized arrays
			if Good:
				finalized_return_array.append(array)
				
		return finalized_return_array
	
	#Step 1 in Isomorphism Algorithm
	def step_1(self, s_graph):
		s_graph_edges = s_graph.get_edges()
		s_graph_nodes = s_graph.get_nodes()
		
		self._matches = []
		
		self._p_a = len(s_graph_nodes)
		self._p_b = len(self._nodes)
		
		self._a = s_graph.get_adjacency()
		self._M = zeros((self._p_a, self._p_b))
		self._M_d = []
		for i in range(self._p_a):
			self._M_d.append(zeros((self._p_a, self._p_b)))
			
		for row in range(self._p_a):
			for col in range(self._p_b):
				if self._nodes[col].contains_subnode(s_graph_nodes[row]):
					self._M[row][col] = 1
					
		self._d = 0
		self._k = -1
		self._H = zeros(self._p_a)
		
		self._H = self._H - 1
		self._F = zeros(self._p_b)
		
		return 2
		
	#A copying matrix for performing deep copies
	#on matrices 
	def copy_matrix(self, matrix, copy_to):
		for row in range(self._p_a):
			for col in range(self._p_b):
				copy_to[row][col] = matrix[row][col]
	
	#Step 2 in the Isomorphism Algorithm
	def step_2(self):
		found_one = False
		for j in range(self._p_b):
			if self._M[self._d][j] == 1 and self._F[j] == 0:
				found_one = True
		
		if found_one:
			self.copy_matrix(self._M, self._M_d[self._d])
			if self._d == 0:
				self._k = self._H[0]
			else:
				self._k = -1
			return 3
		else:	
			return 7
			
	#Step 3 in the Isomorphism Algorithm
	def step_3(self):
		
		self._k += 1
		if self._M[self._d][self._k] == 0 or self._F[self._k] == 1:
			return 3
		for j in range(self._p_b):
			if not j == self._k:
				self._M[self._d][j] = 0
		return 4
	
	#Step 4 in the Isomorphism Algorithm
	def step_4(self):
		if self._d < (self._p_a - 1):
			return 6

		#The actual check for an isomorphism!
		isomorphism = True
		B = dot(self._M,self._adjacency)
		B = B.transpose()
		C = dot(self._M,B)
		C = C.transpose()
		for i in range(self._p_a):
			for j in range(self._p_a):
				if self._a[i][j] == 1:
					if not C[i][j] == 1:
						isomorphism = False
						
		if isomorphism:
			iso = zeros((self._p_a, self._p_b))
			self.copy_matrix(self._M, iso)
			self._matches.append(iso)
		return 5
	
	#Step 5 in the Isomorphism Algorithm
	def step_5(self):
		found_one = False
		for j in range(self._p_b):
			if j > self._k:
				if self._M[self._d][j] == 1 and self._F[j] == 0:
					found_one = True
		
		if not found_one:
			return 7
			
		self.copy_matrix(self._M_d[self._d], self._M)
		return 3
	
	
	#Step 6 in the Isomorphism Algorithm
	def step_6(self):
		self._H[self._d] = self._k
		self._F[self._k] = 1
		self._d += 1
		return 2

	#Step 7 in the Isomorphism Algorithm
	def step_7(self):
		if self._d == 0:
			return -1
		self._F[self._k] = 0
		
		self._d -= 1
		
		self.copy_matrix(self._M_d[self._d], self._M)
		self._k = self._H[self._d]
		return 5
	
	"""-----------------------------------------
	
				End of Algorithm!
	
	-----------------------------------------"""
	
	"""Connect two nodes using an edge"""
	def connect(self, from_node, edge_name, to_node):
		new_edge = Edge(edge_name, from_node, to_node)
		from_node.add_outgoing_edge(new_edge)
		to_node.add_incoming_edge(new_edge)
		self._edges.append(new_edge)
		self.setAdj(from_node, to_node)
		
	"""Make a two way connection (two edges)"""
	def doubleconnect(self, node_1, edge_name, node_2):
		
		#Make the first edge
		edge_1 = Edge(edge_name, node_1, node_2)
		node_1.add_outgoing_edge(edge_1)
		node_2.add_incoming_edge(edge_1)
		self._edges.append(edge_1)
		
		#Make the second edge
		edge_2 = Edge(edge_name, node_2, node_1)
		node_2.add_outgoing_edge(edge_2)
		node_1.add_incoming_edge(edge_2)
		self._edges.append(edge_2)
		
		#Set our adjacencies
		self.setAdj(node_1, node_2)
		self.setAdj(node_2, node_1)
		
	"""Plot our graph in dot format"""
	def plot(self, file_name):
		
		output_file = open("./" + file_name + ".dot", "w")
		output_file.write("digraph " + self._name + " {\n")
		output_file.write("\tnode [ shape = \"record\"]\n")
        
		#Define our nodes
		for node in self._nodes:
			output_file.write("\t" + node.get_name() + " [ label = \"{ " + node.get_name() + " | ")
			
			attr = node.get_attributes()
			if not attr == None:
				for i in attr.keys():
					output_file.write(i + " = " + str(attr[i]) + "\\l")
			output_file.write(" }\" ]\n")
			
		#Write our edges
		for edge in self._edges:
			output_file.write("\t" + edge.get_from_node().get_name() + " -> " + edge.get_to_node().get_name() + " [label=\"" + str(edge.get_name()) + "\"]\n")
		output_file.write("}\n")
		output_file.close()
