"""
StoryNode.py

By: Ben Kybartas
Student ID: 260477933

Represents a Node in our Story "Narrative" Graph.
"""

from Node import *

class StoryNode(Node):
	
	"""
	The Initialization Function
	
	params:
		name = The name of the node
		attributes = The attributes for the node
		target = The target entity for the event
	"""
	def __init__(self, name, attributes, target):
		
		#Set our inputs
		self._name = name
		self._attributes = attributes
		
		#Initially we have no incoming or outgoing edges
		self._incoming_edges = []
		self._outgoing_edges = []
		
		#Set our target
		self._target = target

		#We do not link the nodes to targets
		#and social modifications at start
		self._linked_to_node = None
		self._social_modification = None
	
	"""
	Get Target
	
	Returns the target of our event
	"""
	def get_target(self):
		return self._target
		
	"""
	Set our Target node
	
	params:
		node: The target node
	"""
	def set_target(self, node):
		self._target = node
	
	"""
	Get the social modification from the node
	
	returns the social modification
	"""
	def get_modification(self):
		return self._social_modification
	
	"""
	Set the social modfication for the node
	
	params:
		modification: The social modification we are setting
	"""
	def set_modification(self, modification):
		self._social_modification = modification
		
	"""
	Get the entity we link to
	
	returns the entity object
	"""
	def get_linked_to_node(self):
		return self._linked_to_node
	
	"""
	Set the entity we link to
	
	params:
		node: The node to which we are linking
	"""
	def set_linked_to_node(self, node):
		self._linked_to_node = node

	"""
	def contains_node(self, sub_node):
		for node in self._nodes:
			if node.contains_subnode(sub_node):
				return node
	"""
